package java.lang;

import java.util.WeakHashMap;
import java.lang.ref.WeakReference;
import java.util.concurrent.atomic.AtomicInteger;

public abstract class ClassValue<T>
{
    private static final ClassValue.Entry<?>[] EMPTY_CACHE;
    final int hashCodeForCache;
    private static final AtomicInteger nextHashCode;
    private static final int HASH_INCREMENT = 1640531527;
    static final int HASH_MASK = 1073741823;
    final ClassValue.Identity identity;
    private volatile ClassValue.Version<T> version;
    private static final Object CRITICAL_SECTION;
    
    protected ClassValue() {
        this.hashCodeForCache = (ClassValue.nextHashCode.getAndAdd(1640531527) & 0x3FFFFFFF);
        this.identity = new ClassValue.Identity();
        this.version = new ClassValue.Version(this);
    }
    
    protected abstract T computeValue(Class<?> p0);
    
    public T get(Class<?> clazz) {
        ClassValue.Entry[] cacheCarefully;
        ClassValue.Entry<Object> probeHomeLocation = ClassValue.ClassValueMap.probeHomeLocation(cacheCarefully = getCacheCarefully(clazz), this);
        if (this.match(probeHomeLocation)) {
            return probeHomeLocation.value();
        }
        return this.getFromBackup(cacheCarefully, clazz);
    }
    
    public void remove(Class<?> clazz) {
        getMap(clazz).removeEntry(this);
    }
    
    void put(Class<?> clazz, T t) {
        getMap(clazz).changeEntry(this, t);
    }
    
    private static ClassValue.Entry<?>[] getCacheCarefully(Class<?> clazz) {
        ClassValue.ClassValueMap classValueMap = clazz.classValueMap;
        if (classValueMap == null) {
            return ClassValue.EMPTY_CACHE;
        }
        return classValueMap.getCache();
    }
    
    private T getFromBackup(ClassValue.Entry<?>[] array, Class<?> clazz) {
        ClassValue.Entry<Object> probeBackupLocations = ClassValue.ClassValueMap.probeBackupLocations(array, this);
        if (probeBackupLocations != null) {
            return probeBackupLocations.value();
        }
        return this.getFromHashMap(clazz);
    }
    
    ClassValue.Entry<T> castEntry(ClassValue.Entry<?> entry) {
        return entry;
    }
    
    private T getFromHashMap(Class<?> clazz) {
        ClassValue.ClassValueMap map = getMap(clazz);
        while (true) {
            ClassValue.Entry<Object> entry = map.startEntry(this);
            if (!entry.isPromise()) {
                return entry.value();
            }
            ClassValue.Entry<Object> finishEntry;
            try {
                entry = makeEntry(entry.version(), this.computeValue(clazz));
                entry = map.finishEntry(this, entry);
            }
            finally {
                finishEntry = map.finishEntry(this, entry);
            }
            if (finishEntry != null) {
                return finishEntry.value();
            }
            continue;
        }
    }
    
    boolean match(ClassValue.Entry<?> entry) {
        return entry != null && entry.get() == this.version;
    }
    
    ClassValue.Version<T> version() {
        return this.version;
    }
    
    void bumpVersion() {
        this.version = new ClassValue.Version(this);
    }
    
    private static ClassValue.ClassValueMap getMap(Class<?> clazz) {
        ClassValue.ClassValueMap classValueMap = clazz.classValueMap;
        if (classValueMap != null) {
            return classValueMap;
        }
        return initializeMap(clazz);
    }
    
    private static ClassValue.ClassValueMap initializeMap(Class<?> clazz) {
        ClassValue.ClassValueMap classValueMap;
        synchronized (ClassValue.CRITICAL_SECTION) {
            if ((classValueMap = clazz.classValueMap) == null) {
                classValueMap = (clazz.classValueMap = new ClassValueMap(clazz));
            }
        }
        return classValueMap;
    }
    
    static <T> ClassValue.Entry<T> makeEntry(ClassValue.Version<T> version, T t) {
        return new ClassValue.Entry(version, t);
    }
    
    static {
        EMPTY_CACHE = new ClassValue.Entry[] { null };
        nextHashCode = new AtomicInteger();
        CRITICAL_SECTION = new Object();
    }
    
    static class Identity
    {
    }
    
    static class Version<T>
    {
        private final ClassValue<T> classValue;
        private final ClassValue.Entry<T> promise;
        
        Version(ClassValue<T> classValue) {
            this.promise = new ClassValue.Entry(this);
            this.classValue = classValue;
        }
        
        ClassValue<T> classValue() {
            return this.classValue;
        }
        
        ClassValue.Entry<T> promise() {
            return this.promise;
        }
        
        boolean isLive() {
            return this.classValue.version() == this;
        }
    }
    
    static class Entry<T> extends WeakReference<ClassValue.Version<T>>
    {
        final Object value;
        static final ClassValue.Entry<?> DEAD_ENTRY;
        
        Entry(ClassValue.Version<T> referent, T value) {
            super(referent);
            this.value = value;
        }
        
        private void assertNotPromise() {
            assert !this.isPromise();
            return;
        }
        
        Entry(ClassValue.Version<T> referent) {
            super(referent);
            this.value = this;
        }
        
        T value() {
            this.assertNotPromise();
            return this.value;
        }
        
        boolean isPromise() {
            return this.value == this;
        }
        
        ClassValue.Version<T> version() {
            return (ClassValue.Version)this.get();
        }
        
        ClassValue<T> classValueOrNull() {
            ClassValue.Version<T> version = this.version();
            return (version == null) ? null : version.classValue();
        }
        
        boolean isLive() {
            ClassValue.Version<T> version = this.version();
            if (version == null) {
                return false;
            }
            if (version.isLive()) {
                return true;
            }
            this.clear();
            return false;
        }
        
        ClassValue.Entry<T> refreshVersion(ClassValue.Version<T> version) {
            this.assertNotPromise();
            ClassValue.Entry entry = new ClassValue.Entry(version, this.value);
            this.clear();
            return entry;
        }
        
        static {
            DEAD_ENTRY = new ClassValue.Entry(null, null);
        }
    }
    
    static class ClassValueMap extends WeakHashMap<ClassValue.Identity, ClassValue.Entry<?>>
    {
        private final Class<?> type;
        private ClassValue.Entry<?>[] cacheArray;
        private int cacheLoad;
        private int cacheLoadLimit;
        private static final int INITIAL_ENTRIES = 32;
        private static final int CACHE_LOAD_LIMIT = 67;
        private static final int PROBE_LIMIT = 6;
        
        ClassValueMap(Class<?> type) {
            this.type = type;
            this.sizeCache(32);
        }
        
        ClassValue.Entry<?>[] getCache() {
            return this.cacheArray;
        }
        
        synchronized <T> ClassValue.Entry<T> startEntry(ClassValue<T> classValue) {
            ClassValue.Entry<T> entry = (ClassValue.Entry)this.get(classValue.identity);
            ClassValue.Version<Object> version = classValue.version();
            if (entry == null) {
                ClassValue.Entry<Object> promise = version.promise();
                this.put(classValue.identity, promise);
                return promise;
            }
            if (entry.isPromise()) {
                if (entry.version() != version) {
                    entry = version.promise();
                    this.put(classValue.identity, entry);
                    return entry;
                }
                return entry;
            }
            else {
                if (entry.version() != version) {
                    entry = entry.refreshVersion(version);
                    this.put(classValue.identity, entry);
                }
                this.checkCacheLoad();
                this.addToCache(classValue, entry);
                return entry;
            }
        }
        
        synchronized <T> ClassValue.Entry<T> finishEntry(ClassValue<T> classValue, ClassValue.Entry<T> refreshVersion) {
            ClassValue.Entry<Object> entry = (ClassValue.Entry)this.get(classValue.identity);
            if (refreshVersion == entry) {
                assert !(!refreshVersion.isPromise());
                this.remove(classValue.identity);
                return null;
            }
            else if (entry != null && entry.isPromise() && entry.version() == refreshVersion.version()) {
                ClassValue.Version<Object> version = classValue.version();
                if (refreshVersion.version() != version) {
                    refreshVersion = refreshVersion.refreshVersion(version);
                }
                this.put(classValue.identity, refreshVersion);
                this.checkCacheLoad();
                this.addToCache(classValue, refreshVersion);
                return refreshVersion;
            }
            else {
                return null;
            }
        }
        
        synchronized void removeEntry(ClassValue<?> classValue) {
            ClassValue.Entry<Object> value = (ClassValue.Entry)this.remove(classValue.identity);
            if (value != null) {
                if (value.isPromise()) {
                    this.put(classValue.identity, value);
                }
                else {
                    classValue.bumpVersion();
                    this.removeStaleEntries(classValue);
                }
            }
        }
        
        synchronized <T> void changeEntry(ClassValue<T> classValue, T t) {
            ClassValue.Entry<Object> entry = (ClassValue.Entry)this.get(classValue.identity);
            ClassValue.Version<T> version = classValue.version();
            if (entry != null) {
                if (entry.version() == version && entry.value() == t) {
                    return;
                }
                classValue.bumpVersion();
                this.removeStaleEntries(classValue);
            }
            ClassValue.Entry<T> entry2 = ClassValue.makeEntry(version, t);
            this.put(classValue.identity, entry2);
            this.checkCacheLoad();
            this.addToCache(classValue, entry2);
        }
        
        static ClassValue.Entry<?> loadFromCache(ClassValue.Entry<?>[] array, int n) {
            return array[n & array.length - 1];
        }
        
        static <T> ClassValue.Entry<T> probeHomeLocation(ClassValue.Entry<?>[] array, ClassValue<T> classValue) {
            return classValue.castEntry(loadFromCache(array, classValue.hashCodeForCache));
        }
        
        static <T> ClassValue.Entry<T> probeBackupLocations(ClassValue.Entry<?>[] array, ClassValue<T> classValue) {
            int n = array.length - 1;
            int n2 = classValue.hashCodeForCache & n;
            ClassValue.Entry<?> entry = array[n2];
            if (entry == null) {
                return null;
            }
            int n3 = -1;
            int i = n2 + 1;
            while (i < n2 + 6) {
                ClassValue.Entry<?> entry2 = array[i & n];
                if (entry2 == null) {
                    break;
                }
                if (classValue.match(entry2)) {
                    array[n2] = entry2;
                    if (n3 >= 0) {
                        array[i & n] = ClassValue.Entry.DEAD_ENTRY;
                    }
                    else {
                        n3 = i;
                    }
                    array[n3 & n] = ((entryDislocation(array, n3, entry) < 6) ? entry : ClassValue.Entry.DEAD_ENTRY);
                    return classValue.castEntry(entry2);
                }
                else {
                    if (!entry2.isLive() && n3 < 0) {
                        n3 = i;
                    }
                    ++i;
                }
            }
            return null;
        }
        
        private static int entryDislocation(ClassValue.Entry<?>[] array, int n, ClassValue.Entry<?> entry) {
            ClassValue<Object> classValueOrNull = entry.classValueOrNull();
            if (classValueOrNull == null) {
                return 0;
            }
            return n - classValueOrNull.hashCodeForCache & array.length - 1;
        }
        
        private void sizeCache(int n) {
            assert (n & n - 1) == 0x0;
            this.cacheLoad = 0;
            this.cacheLoadLimit = (int)((double)n * 67.0 / 100.0);
            this.cacheArray = new ClassValue.Entry[n];
            return;
        }
        
        private void checkCacheLoad() {
            if (this.cacheLoad >= this.cacheLoadLimit) {
                this.reduceCacheLoad();
            }
        }
        
        private void reduceCacheLoad() {
            this.removeStaleEntries();
            if (this.cacheLoad < this.cacheLoadLimit) {
                return;
            }
            ClassValue.Entry[] cache = this.getCache();
            if (cache.length > 1073741823) {
                return;
            }
            this.sizeCache(cache.length * 2);
            for (ClassValue.Entry entry : cache) {
                if (entry != null && entry.isLive()) {
                    this.addToCache(entry);
                }
            }
            return;
        }
        
        private void removeStaleEntries(ClassValue.Entry<?>[] array, int i, int n) {
            int n2 = array.length - 1;
            int n3 = 0;
            ClassValue.Entry<?> entry;
            for (; i < i + n; ++i) {
                entry = array[i & n2];
                if (entry != null && !entry.isLive() && (array[i & n2] = this.findReplacement(array, i)) == null) {
                    ++n3;
                }
            }
            this.cacheLoad = Math.max(0, this.cacheLoad - n3);
        }
        
        private ClassValue.Entry<?> findReplacement(ClassValue.Entry<?>[] array, int n) {
            ClassValue.Entry<?> entry = null;
            int n2 = -1;
            int n3 = 0;
            int n4 = array.length - 1;
            int i = n + 1;
            while (i < n + 6) {
                ClassValue.Entry<?> entry2 = array[i & n4];
                if (entry2 == null) {
                    break;
                }
                if (entry2.isLive()) {
                    int entryDislocation = entryDislocation(array, i, entry2);
                    if (entryDislocation != 0) {
                        int n5 = i - entryDislocation;
                        if (n5 <= n) {
                            if (n5 == n) {
                                n2 = 1;
                                n3 = i;
                                entry = entry2;
                            }
                            else if (n2 <= 0) {
                                n2 = 0;
                                n3 = i;
                                entry = entry2;
                            }
                        }
                    }
                }
                ++i;
            }
            if (n2 >= 0) {
                if (array[n3 + 1 & n4] != null) {
                    array[n3 & n4] = ClassValue.Entry.DEAD_ENTRY;
                }
                else {
                    array[n3 & n4] = null;
                    --this.cacheLoad;
                }
            }
            return entry;
        }
        
        private void removeStaleEntries(ClassValue<?> classValue) {
            this.removeStaleEntries(this.getCache(), classValue.hashCodeForCache, 6);
        }
        
        private void removeStaleEntries() {
            ClassValue.Entry[] cache = this.getCache();
            this.removeStaleEntries(cache, 0, cache.length + 6 - 1);
        }
        
        private <T> void addToCache(ClassValue.Entry<T> entry) {
            ClassValue<T> classValueOrNull = entry.classValueOrNull();
            if (classValueOrNull != null) {
                this.addToCache(classValueOrNull, entry);
            }
        }
        
        private <T> void addToCache(ClassValue<T> classValue, ClassValue.Entry<T> entry) {
            ClassValue.Entry[] cache = this.getCache();
            int n = cache.length - 1;
            int n2 = classValue.hashCodeForCache & n;
            ClassValue.Entry<Object> placeInCache = this.placeInCache(cache, n2, entry, false);
            if (placeInCache == null) {
                return;
            }
            int i = n2 - entryDislocation(cache, n2, placeInCache);
            while (i < i + 6) {
                if (this.placeInCache(cache, i & n, placeInCache, true) == null) {
                    return;
                }
                ++i;
            }
            return;
        }
        
        private ClassValue.Entry<?> placeInCache(ClassValue.Entry<?>[] array, int n, ClassValue.Entry<?> entry, boolean b) {
            ClassValue.Entry<Object> overwrittenEntry = this.overwrittenEntry(array[n]);
            if (b && overwrittenEntry != null) {
                return entry;
            }
            array[n] = entry;
            return overwrittenEntry;
        }
        
        private <T> ClassValue.Entry<T> overwrittenEntry(ClassValue.Entry<T> entry) {
            if (entry == null) {
                ++this.cacheLoad;
            }
            else if (entry.isLive()) {
                return entry;
            }
            return null;
        }
    }
}
